"""Tests for the plotting functions."""

import numpy as np
import pytest
import pandas as pd
import matplotlib.pyplot as plt
from typing import Dict, Any, List

from kompot.anndata.functions import compute_differential_abundance, compute_differential_expression
from kompot.plot.volcano import volcano_de, volcano_da, _infer_de_keys, _infer_da_keys
from kompot.plot.heatmap import heatmap, _infer_heatmap_keys


def create_test_anndata(n_cells=100, n_genes=20):
    """Create a test AnnData object."""
    try:
        import anndata
    except ImportError:
        pytest.skip("anndata not installed, skipping test")
        
    np.random.seed(42)
    
    # Create test data
    X = np.random.normal(0, 1, (n_cells, n_genes))
    
    # Create cell groups for testing
    groups = np.array(['A'] * (n_cells // 2) + ['B'] * (n_cells // 2))
    
    # Create embedding
    obsm = {
        'DM_EigenVectors': np.random.normal(0, 1, (n_cells, 10))
    }
    
    # Create observation dataframe
    obs = pd.DataFrame({
        'group': groups
    })
    
    # Create var_names
    var_names = [f'gene_{i}' for i in range(n_genes)]
    
    # Create var DataFrame with var_names as index
    var = pd.DataFrame(index=var_names)
    
    return anndata.AnnData(X=X, obs=obs, var=var, obsm=obsm)


def create_test_data_with_multiple_runs():
    """Create an AnnData object with multiple runs of DA and DE."""
    adata = create_test_anndata()
    
    # Run 1: Differential abundance
    _ = compute_differential_abundance(
        adata,
        groupby='group',
        condition1='A',
        condition2='B',
        result_key='da_run1'
    )
    
    # Run 2: Differential abundance with different parameters
    _ = compute_differential_abundance(
        adata,
        groupby='group',
        condition1='A',
        condition2='B',
        log_fold_change_threshold=2.0,
        result_key='da_run2'
    )
    
    # Run 1: Differential expression
    _ = compute_differential_expression(
        adata,
        groupby='group',
        condition1='A',
        condition2='B',
        result_key='de_run1',
        compute_mahalanobis=False  # Avoid Mahalanobis computation errors in tests
    )
    
    # Run 2: Differential expression with different parameters
    _ = compute_differential_expression(
        adata,
        groupby='group',
        condition1='A',
        condition2='B',
        n_landmarks=20,
        result_key='de_run2',
        compute_mahalanobis=False  # Avoid Mahalanobis computation errors in tests
    )
    
    # Create a global run history (which doesn't happen in the individual function calls)
    from kompot.anndata.functions import run_differential_analysis
    _ = run_differential_analysis(
        adata,
        groupby='group',
        condition1='A',
        condition2='B',
        abundance_key='da_run3',
        expression_key='de_run3',
        generate_html_report=False,
        compute_mahalanobis=False
    )
    
    return adata


class TestKeyInferenceFunctions:
    """Tests for the key inference helper functions."""
    
    def setup_method(self):
        """Set up test data."""
        self.adata = create_test_data_with_multiple_runs()

    def test_infer_de_keys_with_run_id(self):
        """Test DE key inference with specific run_id."""
        # Check if kompot_run_history exists in adata.uns
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
            
        # Get latest run which should be the one with de_run3
        latest_run = self.adata.uns['kompot_latest_run']
        assert 'expression_key' in latest_run
        assert latest_run['expression_key'] == 'de_run3'
        
        # Test inference with latest run (-1)
        lfc_key, score_key = _infer_de_keys(self.adata, run_id=-1)
        assert 'de_run3' in lfc_key, f"Expected 'de_run3' in inferred key, got {lfc_key}"
        assert score_key is not None
        
        # For this test, we can't easily get the specific run_id for de_run1 or de_run2
        # since we don't know the exact ordering in kompot_run_history
        # Instead, we'll just verify that the keys are correctly inferred when
        # explicitly provided.
        
        # Test direct inference without run_id
        de_keys = [k for k in self.adata.var.columns if 'de_run1' in k and 'lfc' in k]
        if de_keys:
            lfc_key, score_key = _infer_de_keys(self.adata, lfc_key=de_keys[0])
            assert 'de_run1' in lfc_key, f"Expected 'de_run1' in inferred key, got {lfc_key}"
    
    def test_infer_da_keys_with_run_id(self):
        """Test DA key inference with specific run_id."""
        # Check if kompot_run_history exists in adata.uns
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
            
        # Get latest run which should be the one with da_run3
        latest_run = self.adata.uns['kompot_latest_run']
        assert 'abundance_key' in latest_run
        assert latest_run['abundance_key'] == 'da_run3'
        
        # Test inference with latest run (-1)
        lfc_key, pval_key = _infer_da_keys(self.adata, run_id=-1)
        assert 'da_run3' in lfc_key, f"Expected 'da_run3' in inferred key, got {lfc_key}"
        assert pval_key is not None
        
        # For this test, we can't easily get the specific run_id for da_run1 or da_run2
        # since we don't know the exact ordering in kompot_run_history
        # Instead, we'll just verify that the keys are correctly inferred when
        # explicitly provided.
        
        # Test direct inference without run_id
        da_keys = [k for k in self.adata.obs.columns if 'da_run1' in k and 'lfc' in k]
        if da_keys:
            lfc_key, pval_key = _infer_da_keys(self.adata, lfc_key=da_keys[0])
            assert 'da_run1' in lfc_key, f"Expected 'da_run1' in inferred key, got {lfc_key}"
    
    def test_infer_heatmap_keys_with_run_id(self):
        """Test heatmap key inference with specific run_id."""
        # Check if kompot_run_history exists in adata.uns
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
            
        # Get latest run which should be the one with de_run3
        latest_run = self.adata.uns['kompot_latest_run']
        assert 'expression_key' in latest_run
        assert latest_run['expression_key'] == 'de_run3'
        
        # Test inference with latest run (-1)
        lfc_key, score_key = _infer_heatmap_keys(self.adata, run_id=-1)
        assert 'de_run3' in lfc_key, f"Expected 'de_run3' in inferred key, got {lfc_key}"
        assert score_key is not None
        
        # For this test, we can't easily get the specific run_id for de_run1 or de_run2
        # since we don't know the exact ordering in kompot_run_history
        # Instead, we'll just verify that the keys are correctly inferred when
        # explicitly provided.
        
        # Test direct inference without run_id
        de_keys = [k for k in self.adata.var.columns if 'de_run1' in k and 'lfc' in k]
        if de_keys:
            lfc_key, score_key = _infer_heatmap_keys(self.adata, lfc_key=de_keys[0])
            assert 'de_run1' in lfc_key, f"Expected 'de_run1' in inferred key, got {lfc_key}"


class TestPlotFunctions:
    """Tests for the plotting functions with run_id parameter."""
    
    def setup_method(self):
        """Set up test data."""
        self.adata = create_test_data_with_multiple_runs()
        plt.clf()  # Clear any existing plots
    
    def teardown_method(self):
        """Clean up after tests."""
        plt.clf()  # Clear any created plots
        plt.close('all')
    
    def test_volcano_de_with_run_id(self):
        """Test volcano_de function with run_id parameter."""
        # Check if run history exists
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
        
        # Test with negative run_id (-1) for latest run with explicit score_key
        # (since Mahalanobis is not computed in our tests)
        self.adata.var['test_score'] = np.random.rand(self.adata.n_vars)
        fig, ax = volcano_de(
            self.adata,
            run_id=-1,
            score_key='test_score',  # Use test_score since mahalanobis is not computed
            return_fig=True
        )
        assert fig is not None
        assert ax is not None
        
        # The variant below should also work by directly providing keys
        # Add a dummy score to test with
        self.adata.var['test_score'] = np.random.rand(self.adata.n_vars)
        
        de_keys = [k for k in self.adata.var.columns if 'de_run1' in k and 'lfc' in k]
        if de_keys:
            fig, ax = volcano_de(
                self.adata,
                lfc_key=de_keys[0],
                score_key='test_score',
                return_fig=True
            )
            assert fig is not None
            assert ax is not None
    
    def test_volcano_da_with_run_id(self):
        """Test volcano_da function with run_id parameter."""
        # Check if run history exists
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
        
        # Test with negative run_id (-1) for latest run
        fig, ax = volcano_da(
            self.adata,
            run_id=-1,
            return_fig=True
        )
        assert fig is not None
        assert ax is not None
        
        # The variant below should also work by directly providing keys
        da_keys_lfc = [k for k in self.adata.obs.columns if 'da_run1' in k and 'lfc' in k]
        da_keys_pval = [k for k in self.adata.obs.columns if 'da_run1' in k and 'pval' in k]
        
        if da_keys_lfc and da_keys_pval:
            fig, ax = volcano_da(
                self.adata,
                lfc_key=da_keys_lfc[0],
                pval_key=da_keys_pval[0],
                return_fig=True
            )
            assert fig is not None
            assert ax is not None
    
    def test_heatmap_with_run_id(self):
        """Test heatmap function with run_id parameter."""
        # Add some data to var that will be used in heatmap
        self.adata.var['test_score'] = np.random.rand(self.adata.n_vars)
        
        # Check if run history exists
        if 'kompot_run_history' not in self.adata.uns:
            pytest.skip("kompot_run_history not found in adata.uns")
            
        # Test with negative run_id (-1) for latest run
        result = heatmap(
            self.adata,
            run_id=-1,
            score_key='test_score',  # Use test_score since mahalanobis might not be computed
            n_top_genes=5,
            return_fig=True
        )
        assert len(result) >= 2
        assert result[0] is not None  # fig
        assert result[1] is not None  # ax
        
        # The variant below should also work by directly providing keys
        de_keys = [k for k in self.adata.var.columns if 'de_run1' in k and 'lfc' in k]
        if de_keys:
            result = heatmap(
                self.adata,
                lfc_key=de_keys[0],
                score_key='test_score',  # Use test_score since mahalanobis might not be computed
                n_top_genes=5,
                return_fig=True
            )
            assert len(result) >= 2
            assert result[0] is not None  # fig
            assert result[1] is not None  # ax